package com.maraujo

import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer

class ForecastEval(A: Array[ArrayBuffer[Double]], B: Array[ArrayBuffer[Double]], T: Array[ArrayBuffer[Double]], rank: Int, initialT: Int) {
  val S = T(0).length-initialT  // number of future steps to predict
  val factors = Array.tabulate(rank) { i =>  new Factor(A(i), B(i), T(i).takeRight(S)) }

  /**
   * Returns the K non-zeros with highest reconstruction score.
   * @param K The number of non-zeros to be returned.
   * @return A sequence of non-zeros with corresponding reconstruction score.
   */
  def topNonZeros(K: Int) : Array[(Double, (Int, Int, Int))] = {
    //val activeSet = initActiveSet(K)
    val activeSet = mutable.TreeSet.empty[(Double, (Int, Int, Int))]
    val (score, elem) = nextBest()
    activeSet += ((score, elem))

    bestValueSum = factors.map(_.getFrontValue).sum
    var iterCounter = 0

    while (activeSet.head._1 < bestValueSum || activeSet.size < K) {
      val (score, elem) = nextBest()

      if (score > activeSet.head._1 || activeSet.size < K) {
        activeSet += ((score, elem))
        if (activeSet.size > K) activeSet.remove(activeSet.firstKey)
      }

      iterCounter += 1
    }

    println("Elements checked: " + iterCounter)

    activeSet.toArray.sortBy(-_._1).map { case (v, (a, b, c)) => (v, (a, b, c+initialT))}
  }

  def nonZerosPerStep(K: Int) : Array[Array[(Double, (Int, Int, Int))]] = {
    val activeSet = Array.fill(S)(mutable.TreeSet.empty[(Double, (Int, Int, Int))])
    activeSet.foreach { set => set.add((0, (-1, -1, -1))) }

    bestValueSum = factors.map(_.getFrontValue).sum
    var iterCounter = 0

    while (activeSet.map(_.head._1).min < bestValueSum || activeSet.map(_.size).min < K) {
      val (score, elem) = nextBest()
      if (score > activeSet(elem._3).head._1 || activeSet(elem._3).size < K) {
        activeSet(elem._3) += ((score, elem))
        if (activeSet(elem._3).size > K) activeSet(elem._3).remove(activeSet(elem._3).firstKey)
      }
    }

    activeSet.map(_.toArray.sortBy(-_._1).map { case (v, (a, b, c)) => (v, (a, b, c+initialT))})
  }

  private var bestValueSum = 0.0

  private def nextBest() : (Double, (Int, Int, Int)) = {
    val maior = factors.map(_.getFrontValue).zipWithIndex.max

    val p = factors(maior._2).nextNonZero()
    val s = getScore(p._2._1, p._2._2, p._2._3)

    bestValueSum -= maior._1
    bestValueSum += factors(maior._2).getFrontValue

    (s, p._2)
  }

  private def getScore(p: (Int, Int, Int)) : Double = factors.map(_.getScore(p)).sum

/*  private def initActiveSet(numPredicted: Int) : mutable.TreeSet[(Double, (Int, Int, Int))] = {
    val activeSet = scala.collection.mutable.TreeSet.empty[(Double, (Int, Int, Int))]

    // get 'numPredicted/rank' from each factor
    for (i <- 0 until rank) {
      for (j <- 0 until numPredicted/rank) {
        val p = factors(i).nextNonZero()
        val s = getScore(p._2._1, p._2._2, p._2._3)
        activeSet += ((s, p._2))
      }
    }

    activeSet
  }*/
}

